@@ -1,13 +1,82 @@
 import os
 import re
+import select
 import sys
-import inspect
 import tempfile
 import unittest.mock
-from subprocess import Popen, PIPE
+from subprocess import Popen, PIPE, TimeoutExpired
 import time
+import hashlib
+import shutil
+from subprocess import Popen, PIPE, CalledProcessError, run
 import signal
 
+DEFAULT_TIMEOUT = 30
+
+
+def run_and_communicate(
+    test_name,
+    args="0",
+    debug_agent_options="",
+):
+    # Prepare command
+    program = "./rocm-debug-agent-test"
+    cmd = [program]
+    if isinstance(args, (list, tuple)):
+        cmd += [str(a) for a in args]
+    else:
+        cmd.append(str(args))
+    with unittest.mock.patch.dict(
+        os.environ,
+        (
+            {"ROCM_DEBUG_AGENT_OPTIONS": debug_agent_options}
+            if debug_agent_options
+            else {}
+        ),
+    ):
+        p = Popen(cmd, stdout=PIPE, stderr=PIPE)
+
+        def handle_failure(reason):
+            print(f"Test {test_name} FAIL: {reason}")
+            p.kill()
+            try:
+                # Kill first, then communicate to flush any remaining output
+                output, err = p.communicate()
+            except Exception:
+                output, err = b"", b""
+            print("rocm-debug-agent test print out.")
+            print(output.decode("utf-8"))
+            print("rocm-debug-agent test error message.")
+            print(err.decode("utf-8"))
+            return None, None, False
+
+        try:
+            output, err = p.communicate(timeout=DEFAULT_TIMEOUT)
+        except TimeoutExpired:
+            return handle_failure("Timeout reached during communicate.")
+        except Exception:
+            return handle_failure("Unknown exception.")
+
+        out_str = output.decode("utf-8")
+        err_str = err.decode("utf-8")
+        return out_str, err_str, True
+
+
+def check_errors(check_list, out_str, err_str):
+    all_strings_found = True
+    for check_str in check_list:
+        if not (check_str.search(err_str)):
+            all_strings_found = False
+            print('"', check_str, '" Not Found in dump.')
+
+    if not all_strings_found:
+        print("rocm-debug-agent test print out.")
+        print(out_str)
+        print("rocm-debug-agent test error message.")
+        print(err_str)
+
+    return all_strings_found
+
 
 def filter_warnings(err_str):
     """Filter out warnings wich are expected on some archs."""
@@ -40,12 +109,11 @@ else:
     os.environ["HSA_TOOLS_LIB"] = "librocm-debug-agent.so.2"
     os.chdir(test_binary_directory)
     # pre test to check if librocm-debug-agent.so.2 can be found
-    p = Popen(["./rocm-debug-agent-test", "0"], stdout=PIPE, stderr=PIPE)
-    output, err = p.communicate()
-    out_str = output.decode("utf-8")
-    err_str = err.decode("utf-8")
+    out_str, err_str, success = run_and_communicate(
+        test_name="0: default", args="0"
+    )
 
-    if filter_warnings(err_str):
+    if success and filter_warnings(err_str):
         print(err_str)
         if '"librocm-debug-agent.so.2" failed to load' in err_str:
             print(
@@ -57,12 +125,16 @@ else:
 # test 0
 def check_test_0():
     print("Starting rocm-debug-agent-test 0")
-    p = Popen(["./rocm-debug-agent-test", "0"], stdout=PIPE, stderr=PIPE)
-    output, err = p.communicate()
-    out_str = output.decode("utf-8")
-    err_str = err.decode("utf-8")
 
-    # Only print but not throw for err_str, since debug build has print out could be ignored
+    out_str, err_str, success = run_and_communicate(
+        test_name="0: default", args="0"
+    )
+
+    if not success:
+        return False
+
+    # Only print but not throw for err_str, since debug build has print
+    # out could be ignored
     if filter_warnings(err_str):
         print(err_str)
 
@@ -92,25 +164,14 @@ def check_test_1():
         ]
     ]
 
-    p = Popen(["./rocm-debug-agent-test", "1"], stdout=PIPE, stderr=PIPE)
-    output, err = p.communicate()
-    out_str = output.decode("utf-8")
-    err_str = err.decode("utf-8")
+    out_str, err_str, success = run_and_communicate(
+        test_name="1: debug_trap", args="1"
+    )
 
-    # check output string
-    all_output_string_found = True
-    for check_str in check_list:
-        if not (check_str.search(err_str)):
-            all_output_string_found = False
-            print('"', check_str, '" Not Found in dump.')
-
-    if not all_output_string_found:
-        print("rocm-debug-agent test print out.")
-        print(out_str)
-        print("rocm-debug-agent test error message.")
-        print(err_str)
-
-    return all_output_string_found
+    if success:
+        return check_errors(check_list, out_str, err_str)
+    else:
+        return False
 
 
 # test 2
@@ -135,35 +196,25 @@ def check_test_2():
             #                  'global_store_dword' # Without precise memory, we can't guarantee that
         ]
     ]
-    p = Popen(["./rocm-debug-agent-test", "2"], stdout=PIPE, stderr=PIPE)
-    output, err = p.communicate()
-    out_str = output.decode("utf-8")
-    err_str = err.decode("utf-8")
 
-    # check output string
-    all_output_string_found = True
-    for check_str in check_list:
-        if not (check_str.search(err_str)):
-            all_output_string_found = False
-            print('"', check_str, '" Not Found in dump.')
-
-    if not all_output_string_found:
-        print("rocm-debug-agent test print out.")
-        print(out_str)
-        print("rocm-debug-agent test error message.")
-        print(err_str)
-
-    return all_output_string_found
+    out_str, err_str, success = run_and_communicate(
+        test_name="2: memory violation", args="2"
+    )
+    if success:
+        return check_errors(check_list, out_str, err_str)
+    else:
+        return False
 
 
 # test 3: snapshot code object on load
 def check_test_3():
     print("Starting rocm-debug-agent test 3")
 
-    p = Popen(["./rocm-debug-agent-test", "3"], stdout=PIPE, stderr=PIPE)
-    output, err = p.communicate()
-    out_str = output.decode("utf-8")
-    err_str = err.decode("utf-8")
+    out_str, err_str, success = run_and_communicate(
+        test_name="3: snapshot code object on load", args="3"
+    )
+    if not success:
+        return False
 
     found_error = False
 
@@ -192,45 +243,95 @@ def check_test_3():
     return not found_error
 
 
+def _file_checksum(path: str, algo: str = "sha256") -> str:
+    h = hashlib.new(algo)
+    with open(path, "rb") as f:
+        for chunk in iter(lambda: f.read(8192), b""):
+            h.update(chunk)
+    return h.hexdigest()
+
+
+def _has_symbol_with_readelf(path: str, symbol: str) -> bool:
+    # Use readelf -s (symbol table). Return True if symbol name appears.
+    try:
+        res = run(["readelf", "-sW", path], stdout=PIPE, stderr=PIPE, check=True)
+        out = res.stdout.decode(errors="ignore")
+        return symbol in out
+    except (CalledProcessError):
+        return False
+
+
 # test 4: save code object on disk
 def check_test_4():
+    if not shutil.which("readelf"):
+        print(
+            "Tool readelf not found, could not run test 4, but don't know if"
+            " that is an error."
+        )
+        unsupported_tests.append(check_test_4)
+        return True
+
     print("Starting rocm-debug-agent test 4")
 
     with tempfile.TemporaryDirectory() as tmpdir:
-        with unittest.mock.patch.dict(
-            os.environ, {"ROCM_DEBUG_AGENT_OPTIONS": f"-p --save-code-objects={tmpdir}"}
-        ):
-
-            p = Popen(["./rocm-debug-agent-test", "4"], stdout=PIPE, stderr=PIPE)
-            p.wait()
+        run_and_communicate(
+            test_name="4: save code objects",
+            args="4",
+            debug_agent_options=f"-p --save-code-objects={tmpdir}",
+        )
 
+        try:
             code_objects = os.listdir(tmpdir)
-            if len(code_objects) == 0:
-                print(f"No code object found in {tmpdir}")
-                return False
+        except FileNotFoundError as e:
+            raise FileNotFoundError(
+                f"Temporary directory not found: {tmpdir}"
+            ) from e
 
-            # There should be 2 code objects which have the same address in
-            # memory and the same size (but might have different load
-            # addressed).  If the name did not have a "N_" prefix (N being a
-            # unique ID), those would be saved with the same file name, so
-            # the second saved code object would override the first one.
-            #
-            # This means that we should see 2 memory code objects:
-            # - 1_memory___PID_offset_OFF_size_SIZE
-            # - 2_memory___PID_offset_OFF_size_SIZE
-            #
-            # Trimming the "N_memory" prefix should give the same value for
-            # both, so a set containing those suffixes should have less
-            # elements than the set of initial names.
-            mem_cos = [co for co in code_objects if "memory___" in co]
-            if len(mem_cos) == len({co.split("memory")[1] for co in mem_cos}):
-                print("Unexpected number of unique code objects")
-                print(
-                    "List of code objects:\n\t{}" "".format("\n\t".join(code_objects))
+        if len(code_objects) == 0:
+            print(f"No code object found in {tmpdir}")
+            return False
+
+        # Filter to files that contain the target symbol in their
+        # symbol table.
+        full_paths = [
+            os.path.join(tmpdir, f)
+            for f in code_objects
+            if os.path.isfile(os.path.join(tmpdir, f))
+        ]
+        with_symbol = []
+        symbol_to_find = "saved_test_kernel"
+        for path in full_paths:
+            if _has_symbol_with_readelf(path, symbol_to_find):
+                with_symbol.append(path)
+
+        if len(with_symbol) != 2:
+            print(
+                f"Expected exactly 2 code objects containing symbol"
+                f" '{symbol_to_find}', found {len(with_symbol)}"
+            )
+            print(
+                "All saved files:\n\t{}".format("\n\t".join(code_objects))
+            )
+            print(
+                "Files with symbol:\n\t{}".format(
+                    "\n\t".join(os.path.basename(p) for p in with_symbol)
                 )
-                return False
+            )
+            return False
 
-            return True
+        # Compare contents via checksum.
+        checksums = [(_file_checksum(p), p) for p in with_symbol]
+        unique_sums = {cs for cs, _ in checksums}
+        if len(unique_sums) != 1:
+            print(
+                "The two code objects containing the symbol do not"
+                " have identical contents"
+            )
+            for cs, pth in checksums:
+                print(f"{os.path.basename(pth)} -> {cs}")
+            return False
+
+        return True
 
 
 # test 5
@@ -251,9 +352,14 @@ def check_test_5():
         with unittest.mock.patch.dict(
             os.environ, {"ROCM_DEBUG_AGENT_OPTIONS": f"-o {tmpdir}/output_log.txt"}
         ):
-            # Start process without capturing output
-            p = Popen(["./rocm-debug-agent-test", "1"])
-            p.wait()  # Wait for the process to complete
+
+            # Start process, ignore what function return since everything
+            # is written to the file
+            run_and_communicate(
+                "5: -o option",
+                args="1",
+                debug_agent_options=f"-o {tmpdir}/output_log.txt",
+            )
 
             # Read the output log
             with open(f"{tmpdir}/output_log.txt", "r") as f:
@@ -283,26 +389,14 @@ def check_test_6():
         ]
     ]
 
-    with unittest.mock.patch.dict(os.environ, {"ROCM_DEBUG_AGENT_OPTIONS": f"-h"}):
-        p = Popen(["./rocm-debug-agent-test", "0"], stdout=PIPE, stderr=PIPE)
-        output, err = p.communicate()
-        out_str = output.decode("utf-8")
-        err_str = err.decode("utf-8")
+    out_str, err_str, success = run_and_communicate(
+        test_name="6: help", args="0", debug_agent_options="-h"
+    )
 
-        # check output string
-        all_output_string_found = True
-        for check_str in check_list:
-            if not (check_str.search(err_str)):
-                all_output_string_found = False
-                print('"', check_str, '" Not Found in dump.')
-
-        if not all_output_string_found:
-            print("rocm-debug-agent test print out.")
-            print(out_str)
-            print("rocm-debug-agent test error message.")
-            print(err_str)
-
-        return all_output_string_found
+    if success:
+        return check_errors(check_list, out_str, err_str)
+    else:
+        return False
 
 
 # test 7
@@ -316,26 +410,14 @@ def check_test_7():
         ]
     ]
 
-    with unittest.mock.patch.dict(os.environ, {"ROCM_DEBUG_AGENT_OPTIONS": f"-l info"}):
-        p = Popen(["./rocm-debug-agent-test", "1"], stdout=PIPE, stderr=PIPE)
-        output, err = p.communicate()
-        out_str = output.decode("utf-8")
-        err_str = err.decode("utf-8")
+    out_str, err_str, success = run_and_communicate(
+        test_name="7: -l option", args="1", debug_agent_options="-l info"
+    )
 
-        # check output string
-        all_output_string_found = True
-        for check_str in check_list:
-            if not (check_str.search(err_str)):
-                all_output_string_found = False
-                print('"', check_str, '" Not Found in dump.')
-
-        if not all_output_string_found:
-            print("rocm-debug-agent test print out.")
-            print(out_str)
-            print("rocm-debug-agent test error message.")
-            print(err_str)
-
-        return all_output_string_found
+    if success:
+        return check_errors(check_list, out_str, err_str)
+    else:
+        return False
 
 
 # test 8
@@ -356,26 +438,14 @@ def check_test_8():
         ]
     ]
 
-    with unittest.mock.patch.dict(os.environ, {"ROCM_DEBUG_AGENT_OPTIONS": f"--all"}):
-        p = Popen(["./rocm-debug-agent-test", "5"], stdout=PIPE, stderr=PIPE)
-        output, err = p.communicate()
-        out_str = output.decode("utf-8")
-        err_str = err.decode("utf-8")
+    out_str, err_str, success = run_and_communicate(
+        test_name="8: --all option", args="5", debug_agent_options="--all"
+    )
 
-        # check output string
-        all_output_string_found = True
-        for check_str in check_list:
-            if not (check_str.search(err_str)):
-                all_output_string_found = False
-                print('"', check_str, '" Not Found in dump.')
-
-        if not all_output_string_found:
-            print("rocm-debug-agent test print out.")
-            print(out_str)
-            print("rocm-debug-agent test error message.")
-            print(err_str)
-
-        return all_output_string_found
+    if success:
+        return check_errors(check_list, out_str, err_str)
+    else:
+        return False
 
 
 # test 9 SIGQUIT
@@ -384,21 +454,105 @@ def check_test_9():
 
     check_list = [
         re.compile(s)
-        for s in ["s0:", "v0:", "Disassembly for function sigquit_kern\\(int\\*\\)"]
+        for s in [
+            "s0:",
+            "v0:",
+            "Disassembly for function sigquit_kern\\(int\\*\\)",
+        ]
     ]
 
+    LOOP_TIMEOUT = DEFAULT_TIMEOUT  # seconds
+
     p = Popen(["./rocm-debug-agent-test", "6"], stdout=PIPE, stderr=PIPE)
-    l = p.stdout.readline()
-    while "Kernel started" not in l.decode("utf-8"):
-        l = p.stdout.readline()
-    os.kill(p.pid, signal.SIGQUIT)
-    time.sleep(1)
+
+    kernel_started = False
+    wave_seen = False
+    timeout_seen = False
+
+    consumed_out = []
+    consumed_err = []
+
+    deadline = time.monotonic() + LOOP_TIMEOUT
+    streams_to_read = [p.stdout, p.stderr]
+
+    while time.monotonic() < deadline and streams_to_read:
+
+        rlist, _, _ = select.select(streams_to_read, [], [], 1)
+        if not rlist:
+            continue
+
+        for r in rlist:
+            line = r.readline()
+
+            if line == b"":
+                # Reading "" means that we reached EOF on this stream.
+                # Remove it from the streams of interest so every stream
+                # can be fully flushed before we exit the loop.
+                del streams_to_read[streams_to_read.index(r)]
+                continue
+
+            s = line.decode("utf-8")
+            if r is p.stdout:
+                consumed_out.append(s)
+            else:
+                consumed_err.append(s)
+
+            if not kernel_started and "Kernel started" in s:
+                kernel_started = True
+                os.kill(p.pid, signal.SIGQUIT)
+                # We give our program 30 secs to start the kernel.
+                # Once we know that the kernel is running, we give it
+                # extra 30 seconds to process SIGQUIT.
+                deadline = deadline + LOOP_TIMEOUT
+
+            if kernel_started:
+                if s.lstrip().startswith(
+                    "Disassembly for function sigquit_kern(int*):"
+                ):
+                    wave_seen = True
+                    break
+            if "Timeout reached. Exiting." in s:
+                timeout_seen = True
+        if wave_seen or timeout_seen:
+            break
+
     p.terminate()
-    output, err = p.communicate()
-    out_str = output.decode("utf-8")
-    err_str = err.decode("utf-8")
+    try:
+        output, err = p.communicate(timeout=3)
+    except TimeoutExpired:
+        print("Timeout reached during final communicate.")
+        output, err = b"", b""
+    except Exception:
+        print("Unexpected exception during final communicate.")
+        output, err = b"", b""
+
+    out_str = "".join(consumed_out) + output.decode("utf-8")
+    err_str = "".join(consumed_err) + err.decode("utf-8")
+
+    if not kernel_started:
+        print("Timeout waiting for 'Kernel started'. Terminating process.")
+        print("rocm-debug-agent test print out.")
+        print(out_str)
+        print("rocm-debug-agent test error message.")
+        print(err_str)
+        return False
+
+    if timeout_seen or not wave_seen:
+        if timeout_seen:
+            print("Timeout reached. Exiting. Failing test.")
+        else:
+            print(
+                (
+                    "Loop timed out without receiving expected message. "
+                    "Failing test."
+                )
+            )
+        print("rocm-debug-agent test print out.")
+        print(out_str)
+        print("rocm-debug-agent test error message.")
+        print(err_str)
+        return False
 
-    # check output string
     all_output_string_found = True
     for check_str in check_list:
         if not (check_str.search(err_str)):
@@ -424,13 +578,15 @@ def check_test_10():
         for s in ["c[gid] = a[gid] + b[gid] + (lds_check[0] >> 32);", "if (gid == 0)"]
     ]
 
-    p_debug = Popen(["./rocm-debug-agent-test", "1"], stdout=PIPE, stderr=PIPE)
-    output, err = p_debug.communicate()
-    err_str_debug = err.decode("utf-8")
+    out_str_debug, err_str_debug, success_debug = run_and_communicate(
+        test_name="10: debug info", args="1"
+    )
+    out_str_no_debug, err_str_no_debug, success_no_debug = run_and_communicate(
+        test_name="10: no debug info", args="7"
+    )
 
-    p_no_debug = Popen(["./rocm-debug-agent-test", "7"], stdout=PIPE, stderr=PIPE)
-    output, err = p_no_debug.communicate()
-    err_str_no_debug = err.decode("utf-8")
+    if not (success_no_debug and success_debug):
+        return False
 
     # check if string is in dissasembly of code with debug info but not in other one
     found_in_debug = False
@@ -446,6 +602,7 @@ def check_test_10():
 
 
 test_success = True
+unsupported_tests = []
 
 for deferred_loading in (None, "1", "0"):
     with unittest.mock.patch.dict("os.environ"):
@@ -474,7 +631,9 @@ for deferred_loading in (None, "1", "0"):
         for i, test in enumerate(test_list, start=0):
             result = test()
             test_success &= result
-            if result:
+            if test in unsupported_tests:
+                print(f"Test {i} UNSUPPORTED")
+            elif result:
                 print(f"Test {i} PASS")
             else:
                 print(f"Test {i} FAIL")
